# canaryd
# File: canaryd/plugin.py
# Desc: functions for handling the various built in state gathering plugins

from __future__ import division

import re
import signal

from distutils.spawn import find_executable
from glob import glob
from os import path

from canaryd.packages import six
from canaryd.packages.importlib import import_module

from canaryd.subprocess import get_command_output
from canaryd.log import logger, print_exception # noqa

PLUGINS = []
NAME_TO_PLUGIN = {}


def underscore(name):
    '''
    Transform CamelCase -> snake_case.
    '''

    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


class NoPluginError(Exception):
    pass


class PluginMeta(type):
    def __init__(cls, name, bases, attrs):
        if name == 'Plugin':
            return

        name = underscore(name)

        plugin_instance = cls()
        plugin_instance.name = name

        PLUGINS.append(plugin_instance)
        NAME_TO_PLUGIN[name] = plugin_instance


@six.add_metaclass(PluginMeta)
class Plugin(object):
    diff_updates = True

    class PrepareError(Exception):
        pass

    class TimeoutError(Exception):
        pass

    def __repr__(self):
        return self.name

    @staticmethod
    def is_change(key, previous_item, item):
        return True

    def get_state(self, settings):
        data = get_command_output(
            self.command,
        )

        return self.parse(data)

    def prepare(self, settings):
        command_bits = self.command.split()
        command = command_bits[0]

        if not find_executable(command):
            raise OSError('Could not find executable: {0}'.format(command))

    def log(self, message):
        message = '[{0}]: {1}'.format(self.name, message)
        logger.debug(message)

    def check_spec_key(self, key, value):
        spec = self.spec[1]

        if key not in spec:
            raise ValueError(
                'Key not found in plugin {0} spec: {1} (value = {2})'.format(
                    self.name,
                    key,
                    value,
                ),
            )

    # Serialisation
    #

    def serialise_value(self, key, value):
        spec = self.spec[1]
        self.check_spec_key(key, value)

        if isinstance(spec[key], set) and isinstance(value, set):
            return list(value)

        return value

    def serialise_item(self, item):
        return dict(
            (key, self.serialise_value(key, value))
            for key, value in six.iteritems(item)
        )

    def serialise_state(self, state):
        return dict(
            (key, self.serialise_item(item))
            for key, item in six.iteritems(state)
        )

    def serialise_changes(self, changes):
        return dict(
            (key, (self.serialise_value(key, item[0]), self.serialise_value(key, item[1])))
            for key, item in six.iteritems(changes)
        )

    # Unserialisation
    #

    def unserialise_value(self, key, value):
        spec = self.spec[1]
        self.check_spec_key(key, value)

        if isinstance(spec[key], set):
            return set(value)

        return value

    def unserialise_item(self, item):
        return dict(
            (key, self.unserialise_value(key, value))
            for key, value in six.iteritems(item)
        )

    def unserialise_state(self, items):
        return dict(
            (key, self.unserialise_item(item))
            for key, item in six.iteritems(items)
        )


def get_plugins():
    '''
    Get the list of installed plugins.
    '''

    module_filenames = glob(path.join(path.dirname(__file__), 'plugins', '*.py'))

    module_names = [
        path.basename(name)[:-3]
        for name in module_filenames
        if not name.endswith('__init__.py')
    ]

    # Import all the modules to populate PLUGINS
    for name in module_names:
        import_module('canaryd.plugins.{0}'.format(name))

    return PLUGINS


def get_plugin_names():
    return six.iterkeys(NAME_TO_PLUGIN)


def get_plugin_by_name(plugin_name):
    if plugin_name not in NAME_TO_PLUGIN:
        raise NoPluginError('Missing plugin: {0}'.format(plugin_name))

    return NAME_TO_PLUGIN[plugin_name]


def prepare_plugin(plugin, settings):
    logger.debug('Preparing plugin: {0}'.format(plugin))

    try:
        plugin.prepare(settings)

    except Plugin.PrepareError as e:
        logger.info('Plugin prepare failed: {0}: {1}'.format(
            plugin.name, e.message,
        ))
        print_exception(debug_only=True)
        return False, e

    except OSError as e:
        logger.info('Plugin command missing/failed: {0}'.format(
            plugin.name,
        ))
        print_exception(debug_only=True)
        return False, e

    except Exception as e:
        logger.warning('Error preparing plugin: {0}: {1}'.format(
            plugin.name,
            e,
        ))
        print_exception()
        return False, e

    return True


def get_and_prepare_working_plugins(settings):
    all_plugins = get_plugins()
    working_plugins = []

    for plugin in all_plugins:
        status = prepare_plugin(plugin, settings)

        if status is True:
            working_plugins.append(plugin)

    logger.info('Loaded plugins: {0}'.format(', '.join([
        plugin.__class__.__name__
        for plugin in working_plugins
    ])))

    return working_plugins


def get_plugin_state(plugin, settings):
    '''
    Gets state output from a single plugin.
    '''

    def _handle_timeout(signum, frame):
        raise plugin.TimeoutError()

    logger.debug('Running plugin: {0}'.format(plugin))

    # Set an alarm: a plugin can only run for MAX half the interval time
    max_time = int(round(settings.collect_interval_s / 2))
    signal.signal(signal.SIGALRM, _handle_timeout)
    signal.alarm(max_time)

    status = False
    state = None

    try:
        state = plugin.get_state(settings)
        status = True

    except plugin.TimeoutError as e:
        logger.warning('Timeout running plugin: {0}'.format(plugin))
        state = e

    except Exception as e:
        logger.warning('Error running plugin: {0}: {1}'.format(plugin, e))
        print_exception()
        state = e

    # Always reset the alarm
    finally:
        signal.alarm(0)

    if not isinstance(state, (dict, list)):
        state = TypeError('Invalid state type: {0}'.format(state))

    return status, state


def get_plugin_states(plugins, settings):
    '''
    Gets state output from multiple plugins.
    '''

    return [
        (plugin, get_plugin_state(plugin, settings))
        for plugin in plugins
    ]



import os
import sys


if os.name == 'posix' and sys.version_info[0] < 3:
    from canaryd.packages.subprocess32 import *  # noqa
else:
    from subprocess import *  # noqa


def get_command_output(command, *args, **kwargs):
    return check_output(  # noqa
        command,
        shell=True,
        close_fds=True,
        *args, **kwargs
    )



import platform
import socket

from distutils.spawn import find_executable
from os import path, sep as os_sep

from canaryd.packages import six

from canaryd.plugin import Plugin
from canaryd.subprocess import CalledProcessError

from .services_util import (
    get_initd_services,
    get_launchd_services,
    get_pid_to_listens,
    get_supervisor_services,
    get_systemd_services,
    get_upstart_services,
)


def update_missing_keys(target, data):
    target.update(dict(
        (key, value)
        for key, value in six.iteritems(data)
        if key not in target
    ))


def check_port(ip_type, host, port):
    # Open our IPv4 or IPv6 socket
    socket_type = socket.AF_INET if ip_type == 'ipv4' else socket.AF_INET6
    sock = socket.socket(socket_type, socket.SOCK_STREAM)
    sock.settimeout(1)

    # If listening everywhere, just try localhost
    if host == '*':
        if ip_type == 'ipv4':
            host = '127.0.0.1'
        else:
            host = '::1'

    try:
        result = sock.connect_ex((host, port))
        return result == 0

    except (socket.error, socket.gaierror):
        pass

    finally:
        sock.close()


class Services(Plugin):
    '''
    The services plugin provides a combined view of "system" services - ie the
    merge of any running launchctl/rcd/initd/upstart/systemd services.
    '''

    spec = ('service', {
        'running': bool,
        'pid': int,
        'enabled': bool,
        'init_system': six.text_type,
        'ports': set((int,)),
        'up_ports': set((int,)),
    })

    @staticmethod
    def prepare(settings):
        pass

    @staticmethod
    def get_state(settings):
        services = {}

        os_type = platform.system().lower()

        if os_type == 'darwin':
            services = get_launchd_services()

        if find_executable('systemctl'):
            update_missing_keys(services, get_systemd_services())

        if find_executable('initctl'):
            update_missing_keys(services, get_upstart_services())

        if path.exists(path.join(os_sep, 'etc', 'init.d')):
            update_missing_keys(
                services,
                get_initd_services(existing_services=services),
            )

        if find_executable('supervisorctl'):
            update_missing_keys(services, get_supervisor_services())

        # Get mapping of PID -> listening ports
        try:
            pid_to_listens = get_pid_to_listens()

        except (CalledProcessError, OSError):
            pass

        else:
            # Augment services with their ports
            for name, data in six.iteritems(services):
                if 'pid' not in data or data['pid'] not in pid_to_listens:
                    data['ports'] = set()
                    data['up_ports'] = set()
                    continue

                data['ports'] = set(
                    port
                    for _, _, port in pid_to_listens[data['pid']]
                )

                data['up_ports'] = set(
                    port for ip_type, host, port in pid_to_listens[data['pid']]
                    if check_port(ip_type, host, port)
                )

        return services

    @staticmethod
    def generate_events(type_, key, data_changes, settings):
        # For new services, no events
        if type_ == 'added':
            return

        # If the script has been removed, resolve any leftover issues and exit
        # (the delete event is still created).
        if type_ == 'deleted':
            yield 'resolved', None, None
            return

        # Track if the service starts/stops
        if 'running' in data_changes:
            _, to_running = data_changes['running']

            if to_running:
                yield 'updated', '{0} started'.format(key), data_changes

            else:
                yield 'updated', '{0} stopped'.format(key), data_changes
                # We stopped the service - so it's ports being up is no longer
                # an issue.
                return

        # No up/down change but PID changed? Service restarted!
        elif 'pid' in data_changes:
            yield 'updated', '{0} restarted'.format(key), data_changes

        # Finally, check the ports!
        if 'up_ports' in data_changes:
            from_ports, to_ports = data_changes['up_ports']

            # No ports up now?
            if from_ports and not to_ports:
                if settings.service_critical:
                    yield (
                        'critical',
                        'All {0} ports down'.format(key),
                        data_changes,
                    )

            # We lost 1+ port, but not all?
            elif from_ports and len(to_ports) < len(from_ports):
                if settings.service_warning:
                    yield (
                        'warning',
                        'Some {0} ports down'.format(key),
                        data_changes,
                    )

            # We have more ports than before? Assume resolved
            # TODO: improve this
            elif to_ports:
                yield 'resolved', 'All {0} ports up'.format(key), data_changes



import re
import socket

from datetime import datetime, timedelta

from canaryd.plugin import Plugin
from canaryd.subprocess import get_command_output

UPTIME_REGEX = re.compile((
    'up\s+(.*?),\s+[0-9]+ '
    'users?,\s+load averages?: '
    '([0-9]+\.[0-9][0-9]),?\s+([0-9]+\.[0-9][0-9]),?\s+([0-9]+\.[0-9][0-9])'
))


def ensure_datetime(datetime_or_string):
    if isinstance(datetime_or_string, datetime):
        return datetime_or_string

    return datetime.strptime(datetime_or_string, '%Y-%m-%dT%H:%M:%S')


def get_uptime():
    data = []
    output = get_command_output('uptime')

    for line in output.splitlines():
        line = line.strip()
        matches = re.search(UPTIME_REGEX, line)

        if matches:
            duration, av1, av5, av15 = matches.groups()

            days = 0
            hours = 0
            mins = 0

            if 'day' in duration:
                match = re.search('([0-9]+)\s+day', duration)
                days = int(match.group(1))

            if ':' in duration:
                match = re.search('([0-9]+):([0-9]+)', duration)
                hours = int(match.group(1))
                mins = int(match.group(2))

            if 'min' in duration:
                match = re.search('([0-9]+)\s+min', duration)
                mins = int(match.group(1))

            up_since = datetime.utcnow() - timedelta(
                days=days, hours=hours, minutes=mins,
            )

            up_since = up_since.replace(
                second=0,
                microsecond=0,
            )

            data.append(('up_since', up_since))

    return data


def get_uname_data(flag):
    output = get_command_output(
        'uname -{0}'.format(flag),
    )

    return output.strip()


class Meta(Plugin):
    spec = ('key', {
        'value': None,
    })

    @staticmethod
    def prepare(settings):
        pass

    @staticmethod
    def get_state(settings):
        data = get_uptime()
        data.append(('hostname', socket.gethostname()))
        data.append(('kernel', get_uname_data('s')))
        data.append(('kernel_release', get_uname_data('r')))
        data.append(('arch', get_uname_data('m')))

        # Nest each value in a dict
        return dict(
            (key, {
                'value': value,
            })
            for key, value in data
        )

    @staticmethod
    def is_change(key, previous_item, item):
        # Check to see if we rebooted, if not there's no change
        if key == 'up_since':
            previous_up = ensure_datetime(previous_item['value'])
            # Account for slight jitter
            previous_up += timedelta(minutes=1)

            up = ensure_datetime(item['value'])

            if (previous_up < up):
                return True
            else:
                return False

        return True

    @staticmethod
    def event_message(type_, key, data_changes):
        if type_ != 'updated':
            return

        if key == 'up_since':
            return 'Server rebooted'

        if key == 'hostname':
            return 'Hostname changed'



from distutils.spawn import find_executable

from canaryd.packages import six

from canaryd.plugin import Plugin

from .packages_util import (
    get_deb_packages,
    get_pkg_packages,
    get_rpm_packages,
)


class Packages(Plugin):
    spec = ('package', {
        'versions': [six.text_type],
        'package_type': six.text_type,
    })

    @staticmethod
    def prepare(settings):
        pass

    @staticmethod
    def get_state(settings):
        packages = {}

        if find_executable('dpkg'):
            packages.update(get_deb_packages())

        if find_executable('rpm'):
            packages.update(get_rpm_packages())

        if find_executable('pkg_info'):
            packages.update(get_pkg_packages())

        return packages



from canaryd.packages import six

from canaryd.plugin import Plugin

# Attempt to import pwd/grp - the users plugin is *nix only
try:
    import pwd
    import grp
except ImportError:
    pwd = None
    grp = None


class Users(Plugin):
    spec = ('user', {
        'group': six.text_type,
        'groups': set((six.text_type,)),
        'home': six.text_type,
        'shell': six.text_type,
    })

    def prepare(self, settings):
        if any(var is None for var in (pwd, grp)):
            raise self.PrepareError('Missing either pwd or group modules.')

    def get_state(self, settings):
        users = {}

        # Get all groups and map by ID -> group
        groups_data = grp.getgrall()

        groups_by_id = dict(
            (group.gr_gid, group)
            for group in groups_data
        )

        # Get basic user data
        users_data = pwd.getpwall()

        for user in users_data:
            users[user.pw_name] = {
                'groups': set(),
                'home': user.pw_dir,
                'shell': user.pw_shell,
            }

            group = groups_by_id.get(user.pw_gid)
            if group:
                users[user.pw_name]['group'] = group.gr_name

        # Loop back through groups and apply any additional to users
        for group in groups_data:
            for username in group.gr_mem:
                if username not in users:
                    continue

                users[username]['groups'].add(group.gr_name)

        return users



from distutils.spawn import find_executable

from canaryd.packages import six

from canaryd.plugin import Plugin

from .containers_util import (
    get_docker_containers,
    get_lxc_containers,
    get_openvz_containers,
    get_virsh_containers,
)


class Containers(Plugin):
    spec = ('container', {
        'runtime': six.text_type,
        'running': bool,
        'pid': int,
        'command': six.text_type,
        'image': six.text_type,
        'id': six.text_type,
        'environment': [six.text_type],
        'names': [six.text_type],
        'ips': [six.text_type],
    })

    @staticmethod
    def prepare(settings):
        pass

    def get_state(self, settings):
        containers = {}

        if find_executable('docker'):
            containers.update(get_docker_containers())

        if find_executable('lxc'):
            containers.update(get_lxc_containers())

        if find_executable('vzlist'):
            containers.update(get_openvz_containers())

        if find_executable('virsh'):
            containers.update(get_virsh_containers())

        return containers
